
# ------------------------------------------------------------------------------
# MODULE        ASCII_ExtendedFormat
# ------------------------------------------------------------------------------
# PURPOSE       Read files in ASCII extended format generated for spectra 
#               measured the CINDI-2 campaign in the frame of the FRM4DOAS project
# ------------------------------------------------------------------------------
# AUTHOR        Caroline FAYT
# COPYRIGHT     Royal Belgian Institute for Space Aeronomy
# CREATION DATE March 2017
#
# MODIFIED
#
# ------------------------------------------------------------------------------

from tools.doas_lib import *
from tools.XML_tools import *

import netCDF4
import re;


dimensions=('num_rec','det_size');
recordKeys=('date','utc time','utc start time','utc end time','viewing elevation angle','viewing azimuth angle','measurement type',
            'total measurement time','total acquisition time (sec)','exposure time (sec)','solar zenith angle','solar azimuth angle');

recordVals=('','','','',0.,0.,0,0.,0.,0.,0.,0.);
parse_filename={'prefix','project','type','institute','station','instr_number','instr_channel','file_version','datetime_start','datetime_end'};

class ExtendedSpectrum :

    def __init__(self,header,wavelength,spectrum):

       self.header=dict(header);                                                # header
       self.wavelength=wavelength;                                              # wavelength calibration
       self.spectrum=spectrum           ;                                       # spectrum

measurementTypes=dict([('offaxis',QDOAS_MeasType_offaxis),         \
                       ('direct sun',QDOAS_MeasType_ds),           \
                       ('zenith',QDOAS_MeasType_zenith),           \
                       ('almucantar',QDOAS_MeasType_almucantar),   \
                       ('horizon',QDOAS_MeasType_horizon),         \
                       ('direct moon',QDOAS_MeasType_moon)])
fillValue=NaN;

def ReadExtendedAsc(ascFile):

    """
    ReadExtendedAsc(ascFile)

    Read file in extended ASCII format used for CINDI2

    Parameters
    ----------

    ascFile    : name of the input file (ASCII format)                          \n
    """

    spe=[];

    if not os.path.isfile(ascFile):
       print('[ReadExtendedAsc] '+ascFile+' not found !!!');
    else:

        f=open(ascFile,'rb');
        detectorSize=0;
        n=0;

        for line in f:

           lowLine=line.lower();
           lowLine=lowLine.decode('utf8').rstrip();

           if '=' in lowLine:

              splittedLine=lowLine.split('=');
              key=splittedLine[0].strip();
              val=splittedLine[1].strip();

              if "size of the detector" in key:

                 detectorSize=int(val);
                 header=[];
                 wavelength=zeros(detectorSize);
                 spectrum=zeros(detectorSize);

              else:

                 for k in recordKeys:
                    if k in key:
                       if k=='date' or 'utc' in k:
                          header.append((k,val));
                       elif k=='measurement type':
                          header.append((k,measurementTypes[val]));
                       else:
                          numstr=re.findall('[+-]?\d+(?:\.\d+)?',val);
                          if len(numstr):
                             header.append((k,float(numstr[0])));

           else:

              numstr=re.findall('[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?',lowLine);
              if len(numstr)==1:
                 spectrum[n]=float(numstr[0]);
                 n=n+1;
              elif len(numstr)==2:
                 wavelength[n]=float(numstr[0]);
                 spectrum[n]=float(numstr[1]);
                 n=n+1;

           if (detectorSize>0) and (n==detectorSize):

              spe.append(ExtendedSpectrum(header,wavelength,spectrum));
              header=[];
              wavelength=zeros(detectorSize);
              spectrum=zeros(detectorSize);
              n=0;

        f.close();

    return(spe);

def ExtractReference(ascFile,refFile,*args):

    """
    ExtractReference(ascFile,refFile)

    Read file in extended ASCII format used for CINDI2 and extract the reference
    spectrum between 11:30 and 11:40 except if a different time range is given

    Parameters
    ----------

    ascFile    : name of the input file (ASCII format)                          \n
    refFile    : the average of zenith spectra found in the given time range (by default [11:30..11:40]
    startTime  : optional, the start time for reference selection if different from 11:30 (11.5)
    endTime    : optional, the end time for reference selection if different from 11:40 (11.6667)
    """
      
    if not os.path.isfile(ascFile):
       print('[ExtractReference] '+ascFile+' not found !!!');
    elif not os.path.isdir(os.path.dirname(refFile)):
       print('[ExtractReference] '+os.path.dirname(refFile)+' not a directory !!!');
    else:

       # Specify the time range for reference selection
       
       startTime=11.5 if len(args)==0 else args[0];
       endTime=11.6667 if len(args)<2 else args[1];

       # Read the ascii file

       spe=ReadExtendedAsc(ascFile);
       
       lastZenith="00:00:00";   # time for the last zenith found before the start time
       nextZenith="00:00:00";   # time of the first zenith found after the end time
       
       lastZenithFound=0;       # 1 if a zenith is found before the start time
       nextZenithFound=0;       # 1 if a zenith is found after the end time
       
       n=0;                     # number of found reference spectra 

       if size(spe)>0:

          detectorSize=size(spe[0].spectrum);

          ref=zeros(detectorSize);
          wve=zeros(detectorSize);

          # Browse spectra

          for item in spe:

             # Find reference in the given time range

             # print(item.header['measurement type'])

             if ((item.header['measurement type']==QDOAS_MeasType_zenith) or \
                 (item.header['viewing elevation angle']>80.)) and \
                 (str2fracTime(item.header['utc time'])>=startTime) and \
                 (str2fracTime(item.header['utc time'])<endTime):

                 ref=ref+item.spectrum;
                 wve=wve+item.wavelength;

                 n=n+1;

             # Find a zenith spectrum before the given time range

             elif (item.header['measurement type']==QDOAS_MeasType_zenith) and (str2fracTime(item.header['utc time'])<startTime):

                 lastZenith=item.header['utc time'];
                 lastZenithFound=1;

             # Find a zenith spectrum after the given time range   

             elif (item.header['measurement type']==QDOAS_MeasType_zenith) and (str2fracTime(item.header['utc time'])>=endTime) and not nextZenithFound:

                 nextZenith=item.header['utc time'];
                 nextZenithFound=1;

          # Reference spectra found in the given time range, save it

          if n>0:

             ref=reshape(ref/float(n),(detectorSize,1));
             wve=reshape(wve/float(n),(detectorSize,1));

             savetxt(refFile,c_[wve,ref],fmt='%12.6f %.14e',newline='\n',delimiter=' ');

          # No reference spectrum found in the given time range

          elif not nextZenithFound and not lastZenithFound:
             print('[ExtractReference] '+ascFile+' No reference spectrum because no zenith found in the file');
          elif not nextZenithFound:
             print('[ExtractReference] '+ascFile+' No reference spectrum because last zenith of the day is '+lastZenith);
          elif not lastZenithFound:
             print('[ExtractReference] '+ascFile+' No reference spectrum because first zenith of the day is '+nextZenith);
          else:
             print('[ExtractReference] '+ascFile+' No reference spectrum in the specified period ('+ \
                                         sprintf('%02d:%02d-%02d:%02d',int(startTime),int((startTime-int(startTime))*60),int(endTime),int((endTime-int(endTime))*60))+ \
                                       ') (previous zenith : '+lastZenith+', next zenith : '+nextZenith);
