
# ------------------------------------------------------------------------------
# MODULE        netCDF_L1
# ------------------------------------------------------------------------------
# PURPOSE       In the frame of the FRM4DOAS project, convert files in netCDF 
#               Supported native formats :
#
#                  extended ASCII format used for CINDI (format='ASCII'),
#                  CCDEEV used at BIRA-IASB (format='CCDEEV')
#                  MFC_BIRA used for MFC formats (binary or STD) pre-converted
#                           by MFC4QDOAS BIRA converter
# ------------------------------------------------------------------------------
# AUTHOR        Caroline FAYT
# COPYRIGHT     Royal Belgian Institute for Space Aeronomy
# CREATION DATE February 2018
#
# MODIFIED
#
# ------------------------------------------------------------------------------

import configparser;

from tools.netCDF_tools import *
from tools.ASCII_ExtendedFormat import *
from tools.uvvis_gb_CCD import *
from tools.MFC_Format import *

netCDF_fillValue=float('NaN');
netCDF_measurementTypes=dict([('offaxis',QDOAS_MeasType_offaxis),         \
                              ('direct sun',QDOAS_MeasType_ds),           \
                              ('zenith',QDOAS_MeasType_zenith),           \
                              ('almucantar',QDOAS_MeasType_almucantar),   \
                              ('horizon',QDOAS_MeasType_horizon),         \
                              ('direct moon',QDOAS_MeasType_moon)])

class netCDF_L1_attributes:

  def __init__(self,instr_number='', \
                    instr_channel='', \
                    instr_type='maxdoas',\
                    institution='', \
                    pi_name='', \
                    pi_email='', \
                    do_name='', \
                    do_email='', \
                    ds_name='', \
                    ds_email='', \
                    station_name='', \
                    instr_latitude='0.', \
                    instr_longitude='0.', \
                    instr_altitude_asl='0.', \
                    instr_altitude_agl='0.'):

       # Instrument

       self.instr_number=instr_number;               # instrument number
       self.instr_channel=instr_channel;             # channel number : 1, 2, 3 (according to metadata)
       self.instr_type=instr_type;                   # maxdoas, zenith, mobile...

       # data_submitter

       self.institution=institution;                 # the name of the institute
       self.pi_name=pi_name;                         # the name of the pi
       self.pi_email=pi_email;                       # the email of the pi
       self.do_name=do_name;                         # the name of the data creator
       self.do_email=do_email;                       # the email of the data creator
       self.ds_name=ds_name;                         # the name of the data submitter
       self.ds_email=ds_email;                       # the email of the data submitter

       # instrument location
       
       self.station_name=station_name;               # the name of the station (should come from a table)
       self.instr_latitude=instr_latitude;           # the latitude of the instrument
       self.instr_longitude=instr_longitude;         # the longitude of the instrument
       self.instr_altitude_asl=instr_altitude_asl;   # the altitude of the instrument above sea level
       self.instr_altitude_agl=instr_altitude_agl;   # the altitude of the instrument above ground level

       self.prefix='ESA';                            # here, the project financier
       self.project='FRM4DOAS';                      # could be changed in the future

  def LoadFromIni(self,configFile,expectedSections=[]):

     config=configparser.ConfigParser()
     functionName='LoadFromIni';

     if not os.path.isfile(configFile):
       print('['+functionName+'] '+configFile+' not found !!!');
     else:

       try:

           config.read(configFile);
           configSections=config.sections();

           if not len(configSections):
             print('['+functionName+'] no sections found in '+configFile);
           else:
             sectionsToFind=expectedSections if len(expectedSections) else configSections;

             for section in sectionsToFind :
          
               for item in config[section]:
                 setattr(self,item,config[section][item]);
       
       except configparser.Error :
           print('['+functionName+'] '+configFile+' can not be loaded !!!');

       # TODO : check the different values are not empty and valid
       #
       #        instrument number, channel
       #        instrument type from metadata
       #        names and emails of pi,do,ds from a table
       #        station name from a table
       #        latitude,longitude within a radius of x km
       #        values of the altitudes w.r.t. the altitude of the station
       

class netCDF_L1_metadata :
         
  def __init__(self,detsize,nrecords):

     if (detsize>0) or (nrecords>0):

         self.utTime=zeros((nrecords,7),dtype=int16);
         self.utStart=zeros((nrecords,7),dtype=int16);
         self.utEnd=zeros((nrecords,7),dtype=int16);
     
         self.viewingElevation=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.viewingAzimuth=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.solarZenithAngle=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.solarAzimuthAngle=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.exposureTime=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.measureType=ones((nrecords,),dtype=int16)*(-1);

         self.totalAcqTime=ones((nrecords,),dtype=float32)*netCDF_fillValue;
         self.totalMeasTime=ones((nrecords,),dtype=float32)*netCDF_fillValue;
     
class netCDF_L1_measurements :
         
  def __init__(self,detsize,nrecords):

     if (detsize>0) or (nrecords>0):

         self.wavelength=ones((nrecords,detsize),dtype=float32)*netCDF_fillValue;
         self.spectra=ones((nrecords,detsize),dtype=float32)*netCDF_fillValue;
         self.errors=ones((nrecords,detsize),dtype=float32)*netCDF_fillValue;
         self.pixelQualityFlags=ones((nrecords,detsize),dtype=int16);

class netCDF_L1_records :

  def __init__(self,detsize,nrecords):

     self.detsize=detsize;
     self.nrecords=nrecords;
     self.datesize=7;

     if (detsize>0) or (nrecords>0):
         
         self.metadata=netCDF_L1_metadata(detsize,nrecords);
         self.measurements=netCDF_L1_measurements(detsize,nrecords);

class netCDF_L1_tp :
           
  def __init__(self,tpFile):
    
     self.altitude_levels=[];
     self.meteo_time=[];
     self.pressure=[];
     self.temperature=[];

     self.nlevels=0;
     self.ntime=0;
    
     if len(tpFile):
          
       if not os.path.isfile(tpFile):
         print('['+functionName+'] '+tpFile+' not found !!!');
       else:  
          
         tp=loadtxt(tpFile);

         self.nlevels=size(tp,0);
         self.ntime=5;   # could be changed later

         self.meteo_time[:]=array([0.,6.,12.,18.,24.]);
        
         self.altitude_levels=tp[:,0];
         self.pressure=zeros((self.nlevels,self.ntime));
         self.temperature=zeros((self.nlevels,self.ntime));

         for i in arange(0,self.ntime):

           self.pressure[:,i]=tp[:,2];
           self.temperature[:,i]=tp[:,3];
        
def netCDF_L1_LoadfromASCII(ascFile):

    """
    L1=netCDF_L1_LoadfromASCII(ascFile)
    Load Level-1 radiance spectra from a file in ASCII extended format

    Parameters
    ----------

    ascFile : the name of the file to read

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, ReadExtendedAsc (from ASCII_ExtendedFormat)
    """

    functionName="netCDF_LoadL1_fromASCII"
    L1=netCDF_L1_records(0,0);

    if not os.path.isfile(ascFile):
      print('['+functionName+'] '+ascFile+' not found !!!');
    else:

      data=ReadExtendedAsc(ascFile);
      
      nrecords=len(data);
      detsize=0 if nrecords==0 else len(data[0].spectrum);

      if (detsize==0) or (nrecords==0):
        print('['+functionName+'] '+ascFile+' is empty !!!');
      else:
          
        L1=netCDF_L1_records(detsize,nrecords);

        for i in arange(0,nrecords):

          header=data[i].header;

          L1.metadata.viewingElevation[i]=header['viewing elevation angle'];
          L1.metadata.viewingAzimuth[i]=header['viewing azimuth angle'];
          L1.metadata.solarZenithAngle[i]=header['solar zenith angle'];
          
          L1.metadata.solarAzimuthAngle[i]=header['solar azimuth angle'];
          L1.metadata.exposureTime[i]=header['exposure time (sec)'];
          L1.metadata.measureType[i]=header['measurement type'];
          L1.metadata.totalAcqTime[i]=header['total acquisition time (sec)'];
          L1.metadata.totalMeasTime[i]=header['total measurement time'];

          L1.measurements.wavelength[i,:]=data[i].wavelength;
          L1.measurements.spectra[i,:]=data[i].spectrum;

    return L1;

def netCDF_L1_LoadfromCCDEEV(dscFile,clbFile=''):

    """
    L1=netCDF_L1_LoadfromCCDEEV(dscFile,[clbFile])

    Load Level-1 radiance spectra from a file in CCD-EEV format
    This format is used at BIRA-IASB
    For FRM4DOAS, only DSC files (dark current and straylight corrected spectra should be used)

    Parameters
    ----------

    dscFile : the name of the file to read
    clbFile : optional but recommended, the wavelength calibration to associate to the file

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, CCD_ReadSpe (from uvvis_gb_CCD)
    """               

    functionName="netCDF_L1_LoadfromCCDEEV"
    L1=netCDF_L1_records(0,0);
    clb=[];
    if len(clbFile):
       clb=netCDF_L1_LoadCalib(clbFile,commentChar=';');
    
    if not os.path.isfile(dscFile):
      print('['+functionName+'] '+dscFile+' not found !!!');
    else :
      
      data=CCD_ReadSpe(dscFile,0,0,0,0,0);

      nrecords=len(data);
      detsize=0 if nrecords==0 else len(data[0].spectrum);

      if (detsize==0) or (nrecords==0):
        print('['+functionName+'] '+dscFile+' is empty !!!');
      elif (len(clb)>0) and (len(clb)!=detsize):
        print('['+functionName+'] calibration '+clbFile+' should have the same number of points as spectra !!!');
      else:
          
        L1=netCDF_L1_records(detsize,nrecords);

        for i in arange(0,nrecords):

          header=data[i].header;
          measType=header['measureType'];

          if (measType==CCD_MT_offaxis) or (measType==CCD_MT_azimuth) :
             measType=CCD_MT_offaxis;
          elif (measType==CCD_MT_ds) or (measType==CCD_MT_almucantar) or (measType==CCD_MT_horizon) :
             measType=measType;
          else:
             measType=CCD_MT_zenith;

          trackerFlag=(header['targetElevation']==header['targetAzimuth']) and (header['brusagElevation']!=-1.) and (header['brusagAzimuth']!=-1.);
  
          L1.metadata.utTime[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['ti_hour'],header['ti_min'],header['ti_sec'],0],dtype=int16);
          L1.metadata.utStart[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['start_ti_hour'],header['start_ti_min'],header['start_ti_sec'],0],dtype=int16);
          L1.metadata.utEnd[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['end_ti_hour'],header['end_ti_min'],header['end_ti_sec'],0],dtype=int16);

          L1.metadata.viewingElevation[i]=header['targetElevation'] if not trackerFlag else floor(header['brusagElevation']+0.5);
          L1.metadata.viewingAzimuth[i]=header['targetAzimuth'] if not trackerFlag else floor(header['brusagAzimuth']+0.5);
          L1.metadata.solarZenithAngle[i]=header['Zm'];
          L1.metadata.solarAzimuthAngle[i]=header['Azimuth']; 
          
          L1.metadata.exposureTime[i]=header['exposureTime'];
          L1.metadata.measureType[i]=measType;
          L1.metadata.totalAcqTime[i]=float(header['nacc']*header['exposureTime']);
          L1.metadata.totalMeasTime[i]=float((header['end_ti_hour']*3600.+header['end_ti_min']*60.+header['end_ti_sec'])-   \
                                             (header['start_ti_hour']*3600.+header['start_ti_min']*60.+header['start_ti_sec']));

          if len(clb):
            L1.measurements.wavelength[i,:]=clb;
          L1.measurements.spectra[i,:]=reshape(data[i].spectrum,(detsize,));

    return L1;
                 
def netCDF_L1_LoadfromMFC_BIRA(mfcFile,clbFile='',commentChar='#'):
               
    """
    L1=netCDF_L1_LoadfromMFC_BIRA(mfcFile,[clbFile])

    Load Level-1 radiance spectra from a file in MFC BIRA-IASB format
    The format is proper to BIRA-IASB and is generated by a converter (MFC4QDOAS_Converter) from original MFC files (binary or std)
    produced by the acquisition S/W DOASIS.  In the BIRA-IASB, all files what ever the spectrum type (dark current, offset and observations)
    are saved in daily files.  The function MFC_BIRA_ReadSpe (such as QDOAS) reads and average dark current and offset and apply
    the correction to observations.

    Parameters
    ----------

    mfcFile     : the name of the file to read (MFC BIRA-IASB)
    clbFile     : optional but recommended, the wavelength calibration to associate to the file
    commentChar : the character for comments used in clbFile

    Returns
    -------

    L1 : a netCDF_L1_records structure

    See also
    --------

    netCDF_L1_records class, MFC_BIRA_ReadSpe (from MFC_format)
    """               

    functionName="netCDF_L1_LoadfromMFC_BIRA"
    L1=netCDF_L1_records(0,0);
    clb=[];
    if len(clbFile):
       clb=netCDF_L1_LoadCalib(clbFile,commentChar=commentChar);
    
    if not os.path.isfile(mfcFile):
      print('['+functionName+'] '+mfcFile+' not found !!!');
    else :
      
      data=MFC_BIRA_ReadSpe(mfcFile,1);

      nrecords=len(data);
      detsize=0 if nrecords==0 else len(data[0].spectrumCorrected);

      if (detsize==0) or (nrecords==0):
        print('['+functionName+'] '+mfcFile+' is empty !!!');
      elif (len(clb)>0) and (len(clb)!=detsize):
        print('['+functionName+'] calibration '+clbFile+' should have the same number of points as spectra !!!');
      else:
          
        L1=netCDF_L1_records(detsize,nrecords);

        for i in arange(0,nrecords):

          header=data[i].header;
          measType=header['measurementType'];

          if (measType==CCD_MT_offaxis) or (measType==CCD_MT_azimuth) :
             measType=CCD_MT_offaxis;
          elif (measType==CCD_MT_ds) or (measType==CCD_MT_almucantar) or (measType==CCD_MT_horizon) :
             measType=measType;
          else:
             measType=CCD_MT_zenith;
             
          L1.metadata.utStart[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['start_ti_hour'],header['start_ti_min'],header['start_ti_sec'],0],dtype=int16);
          L1.metadata.utEnd[i,:]=array([header['da_year'],header['da_month'],header['da_day'],header['end_ti_hour'],header['end_ti_min'],header['end_ti_sec'],0],dtype=int16);

          dt=datevec((datenum(header['da_year'],header['da_month'],header['da_day'],header['start_ti_hour'],header['start_ti_min'],header['start_ti_sec'])+ \
                     datenum(header['da_year'],header['da_month'],header['da_day'],header['end_ti_hour'],header['end_ti_min'],header['end_ti_sec']))*0.5);

          L1.metadata.utTime[i,0:6]=array(dt,dtype=int16);
          L1.metadata.utTime[i,6]=0;

          L1.metadata.viewingElevation[i]=header['elevationAngle'];
          L1.metadata.viewingAzimuth[i]=header['azimuthAngle'];
          L1.metadata.solarZenithAngle[i]=fillValue;    # todo : should be added by the MFC4QDOAS converter
          L1.metadata.solarAzimuthAngle[i]=fillValue;   # todo : should be added by the MFC4QDOAS converter header['Azimuth'];
          L1.metadata.exposureTime[i]=header['exposureTime'];
          L1.metadata.measureType[i]=measType;
          L1.metadata.totalAcqTime[i]=float(header['scansNumber']*header['exposureTime']);
          L1.metadata.totalMeasTime[i]=float((header['end_ti_hour']*3600.+header['end_ti_min']*60.+header['end_ti_sec'])-   \
                                             (header['start_ti_hour']*3600.+header['start_ti_min']*60.+header['start_ti_sec']));

          if len(clb):
            L1.measurements.wavelength[i,:]=clb;
          L1.measurements.spectra[i,:]=reshape(data[i].spectrumCorrected,(detsize,));

    return L1;
                 
def netCDF_L1_LoadCalib(clbFile,commentChar='#'):

    """
    netCDF_L1_LoadCalib(clbFile,[commentChar])

    Load the calibration file to associate to L1 radiance spectra
    It should be a ASCII file.  The character to use for comments should be provided to loadtxt function if it is not the python one (#)
    
    Parameters
    ----------

    clbFile     : the name of the calibration file
    commentChar : optional, comment character to pass to loadtxt

    Returns
    -------

    the wavelengthcalibration
    """

    # to do : check the wavelength coverage ?

    clb=[];

    functionName="netCDF_L1_LoadCalib";
    if len(clbFile)>0:
       
       if not os.path.isfile(clbFile):
          print('['+functionName+'] '+clbFile+' not found !!!');
       else:
           
         clb=loadtxt(clbFile,comments=commentChar);
         if (clb.ndim>1):
           clb=clb[:,0];

    return clb;

def netCDF_L1_LoadRef(refFile,commentChar='#'):

    """
    netCDF_L1_LoadRef(refFile,[commentChar])

    Load the reference spectrum that will be used for the wavelength calibration
    It should be a two-columns ASCII file.  The character to use for comments should be provided to loadtxt function if it is not the python one (#)
    
    Parameters
    ----------

    refFile     : the name of the reference file
    commentChar : optional, comment character to pass to loadtxt

    Returns
    -------

    the reference calibration and spectrum
    """

    # to do : check the wavelength coverage and the S/N ratio of the reference ?

    ref=[];

    functionName="netCDF_L1_LoadRef";
    if len(refFile)>0:
       
       if not os.path.isfile(refFile):
          print('['+functionName+'] '+refFile+' not found !!!');
       else:
           
         ref=loadtxt(refFile,comments=commentChar);
         if (ref.ndim<2):
            print('['+functionName+'] '+refFile+' should contain at least two columns');

    return ref;        
            
def netCDF_L1_LoadSlitFunction(slfFile,wavelength=0.,commentChar='#'):

    """
    netCDF_L1_LoadSlitFunction(slfFile,[wavelength, commentChar])

    Load the slit funtion that will be used for the wavelength calibration
    It should be a multiple columns ASCII file.  The character to use for comments should be provided to loadtxt function if it is not the python one (#)
    For tables, the first line should contain the different wavelengths at which the slit function have been measured.
    If only one slit function is given, the wavelength should be provided
    
    Parameters
    ----------

    slfFile     : the name of the slit function file
    wavelength  : if only one line shape is given, it is the wavelength at wich the slit function has been measured
    commentChar : optional, comment character to pass to loadtxt

    Returns
    -------

    the table of slit functions
    """

    # to do : check the wavelength calibration (should be relative)
    #         check the slit functions ??? estimation of the FWHM ?
    #         check wavelengths at which the slit function has been measured
    #         allow wavelength dependent slit functions (wavelength dependend gaussian file produced by QDOAS ?)
    
    slf=[];
    slitX=[];
    slitY=[];
    slit=[];

    functionName="netCDF_L1_LoadSlitFunction";
    if len(slfFile)>0:
       
       if not os.path.isfile(slfFile):
          print('['+functionName+'] '+slfFile+' not found !!!');
       else:
           
         slf=loadtxt(slfFile,comments=commentChar);
         if (slf.ndim<2):
            print('['+functionName+'] '+slfFile+' should contain at least two columns');
         elif size(slf,1)==2 and wavelength<300.:
           print('['+functionName+'] wavelength at which the slit function has been measured not specified');
         elif size(slf,1)==2:

             slitX=slf[:,0];
             slitY=array([wavelength]);
             slit=slf[:,1];
         
         else:  # table of slit functions

             slitX=slf[1:size(slf,0),0];
             slitY=slf[0,1:size(slf,1)];
             slit=slf[1:size(slf,0),1:size(slf,1)]; # measured slit functions

    return (slitX,slitY,slit);

def netCDF_L1_BuildFileName(ncPath,                          \
                            startDatetime,endDatetime,       \
                            prefix="ESA",                    \
                            project="FRM4DOAS",              \
                            institution="BIRA.IASB",         \
                            station_name="HARESTUA",         \
                            instr_number="004",              \
                            instr_channel="1",               \
                            file_type="L1",                  \
                            file_version="001"):

    """
    netCDF_L1_BuildFileName(ncPath,startDatetime,endDatetime,[prefix,project,institution,station_name,instr_number,instr_channel,file_typefile_version)

    Build the name of the L1 radiance file according to conventions predefined by the current project.
    - is the field separator

    Parameters
    ----------

    ncPath            : the path in which to create the file name
    startDatetime     : the date and time of the first measurement (should be a yyyy,mm,dd,hh,mm,ss,ms array)
    endDatetime       : the date and time of the last measurement (should be a yyyy,mm,dd,hh,mm,ss,ms array)
    prefix            : optional, the prefix of the file name (should be ESA for ESA projects)
    project           : optional, the name of the project (should be FRM4DOAS for FRM4DOAS)
    institution       : optional, the name of the institute
    station_name      : optional, the name of the station
    instr_number      : optional, the instrument number (should be unique)
    instr_channel     : optional, the channel number
    file_type         : optional, the file type (should be L1 for L1 radiance files)
    file_version      : optional, the file version

    Returns
    -------

    the name of the netCDF file built with input fields
    """

    function_name="netCDF_L1_BuildFileName"
    ncFile="";

  # todo : check the format of startDatetime and endDatetime
  #        check that the institution in data base
  #        check that the station name in the data base
  #        according to the instrument number, check the consistency with instrument channel, institution and station

    if not os.path.isdir(ncPath):
      print('['+function_name+'] '+ncPath+' not a valid directory');
    elif len(prefix)==0 or len(project)==0 or len(institution)==0 or len(station_name)==0 or len(instr_number)==0 or len(instr_channel)==0 or len(file_type)==0 or len(file_version)==0:
      print('['+function_name+'] one of the input arguments is empty');
    else:
    
      d1='%d%02d%02d' % (startDatetime[0],startDatetime[1],startDatetime[2]);
      d2='%d%02d%02d' % (endDatetime[0],endDatetime[1],endDatetime[2]);
      t1='%02d%02d%02d' % (startDatetime[3],startDatetime[4],startDatetime[5]);
      t2='%02d%02d%02d' % (endDatetime[3],endDatetime[4],endDatetime[5]);
              
      ncFile=ncPath+"/"+prefix+"-"+project+"-"+file_type+"-"+institution+"-"+station_name+"-"+instr_number+"-"+instr_channel+"-"+d1+'T'+t1+"Z-"+d2+'T'+t2+"Z-fv"+file_version+".nc";

    return ncFile;

def netCDF_L1_Convert(cfgFile,inpFile,ncPath,refFile='',slfFile='',tpFile='',clbFile='',file_version='001',slfWavelength=0.,file_format='ASCII',commentChar='#',solarAzimuthCorrection=0.,viewingAzimuthCorrection=0.):

    """
    Converter_asc_to_netCDF(inpFile,ncFile)

    Read file in extended ASCII format used for CINDI2 and convert it into the netCDF file

    Parameters
    ----------

    cfgFile                  : ini file with some (key,value) to use as file attributes                                     \n
    inpFile                  : name of the input file (ASCII format)                                                        \n
    ncPath                   : name of the output path                                                                      \n
    refFile                  : optional, reference file for QDOAS calibration as keydata                                    \n
    slfFile                  : optional, slit function file provided by the data submitter as keydata                       \n
    tpFile                   : optional, temperature/profile file                                                           \n
    clbFile                  : optional, wavelength calibration file                                                        \n
    file_version             : optional, the file version on three digits                                                   \n
    slfWavelength            : optional, for 1D slit function, the wavelength at which the slit function has been measured  \n
    file_format              : the native file format (by default, ASCII but CCEDDEV and MFC-BIRA supported                 \n
    commentChar              : optional, the character to use for comment in clbFile, refFile, slfFile                      \n
    solarAzimuthCorrection   : optional, the azimuth correction to apply to solar azimuth angle (+ is applied)              \n
    viewingAzimuthCorrection : optional, the azimuth correction to apply to viewing azimuth angle (+ is applied)            \n
    

    TODO
    ----

    check arguments ??? for example : valid value for format

    """

    functionName="netCDF_L1_Convert"
    ncFile=""

    if not os.path.isfile(cfgFile):
       print('['+functionName+']'+cfgFile+' not found !!!');
    elif not os.path.isfile(inpFile):
       print('['+functionName+']'+inpFile+' not found !!!');
    elif not os.path.isdir(ncPath):
       print('['+functionName+']'+ncPath+' not a directory !!!');
    else:

       # Load input files

       config=netCDF_L1_attributes();  # Initialize attributes

       try:
         
        config.LoadFromIni(cfgFile);                                             # Load attributes from config file

        if (file_format=='CCDEEV'):
           L1_data=netCDF_L1_LoadfromCCDEEV(inpFile,clbFile);
        elif (file_format=='ASCII'):
           L1_data=netCDF_L1_LoadfromASCII(inpFile);                             # Load data from L1 in extended ASCII format
        elif (file_format=='MFC_BIRA'):
           L1_data=netCDF_L1_LoadfromMFC_BIRA(inpFile,clbFile,commentChar=commentChar);
        else:
           print('['+functionName+'] Unknown format '+file_format+' !!!');

        tp=netCDF_L1_tp(tpFile);                                                 # Load temperature/pressure profiles

        ref=netCDF_L1_LoadRef(refFile,commentChar=commentChar);                                          # Load reference file if any
        (slfX,slfY,slf)=netCDF_L1_LoadSlitFunction(slfFile,slfWavelength,commentChar=commentChar);       # Load slit function if any

       except BaseException as error:
        print('['+functionName+'] returned fatal error : {}'.format(error))
        return;

       nrecords=L1_data.nrecords;
       detsize=L1_data.detsize;
       refSize=shape(ref);
       slfSize=shape(slf);
          
       if nrecords==0 or detsize==0:
          print('['+functionName+']'+inpFile+' is empty');
       elif refSize[0]>0 and refSize[0]!=detsize:
          print('['+functionName+'] Spectra and reference do not have the same size');
       else :

          # Build filename (should be something like ESA-FRM4DOAS-L1-BIRA.IASB-HARESTUA-004-1-20171208T083023Z-20171208T151709Z-fv001.nc)

          ncFile=netCDF_L1_BuildFileName(ncPath,                                \
                                         L1_data.metadata.utTime[0,:],          \
                                         L1_data.metadata.utTime[nrecords-1,:], \
                                         prefix=config.prefix,                  \
                                         project=config.project,                \
                                         institution=config.institution,        \
                                         station_name=config.station_name,      \
                                         instr_number=config.instr_number,      \
                                         instr_channel=config.instr_channel,    \
                                         file_type="L1",                        \
                                         file_version=file_version); 

          # Open file

          try:
             nc_root=netCDF4.Dataset(ncFile, 'w', format='NETCDF4');
          except:
             print('['+functionName+'] Can not open '+ncFile+' in writing mode !!!');
             return;

          # Set attributes

          nc_root.Conventions="CF-1.6";
          nc_root.title='Level-1 data for the FRM4DOAS MAXDOAS processor';
          nc_root.institution=config.institution;
          nc_root.source='MAXDOAS calibrated radiances';

          nc_root.pi_name=config.pi_name;
          nc_root.pi_email=config.pi_email;
          nc_root.do_name=config.do_name;
          nc_root.do_email=config.do_email;
          nc_root.ds_name=config.ds_name;
          nc_root.ds_email=config.ds_email;

          nc_root.instrument_number=config.instr_number;
          nc_root.instrument_channel=config.instr_channel;
          nc_root.instrument_type=config.instr_type;

          # Create dimensions

          nc_nrec=nc_root.createDimension('num_rec',nrecords);
          nc_detsize=nc_root.createDimension('det_size',detsize);
          nc_profiles=nc_root.createDimension('meteo_levels_size',tp.nlevels);
          nc_meteotime=nc_root.createDimension('meteo_time_size',tp.ntime);
          nc_onedimension=nc_root.createDimension('dim1_size',1);
          nc_datesize=nc_root.createDimension('datetime_size',7);

          if slfSize[0]>0 and slfSize[1]>0:

             nc_slf_x=nc_root.createDimension('slit_dimx',slfSize[0]);
             nc_slf_y=nc_root.createDimension('slit_dimy',slfSize[1]);

          # Create groups

          nc_instrument=nc_root.createGroup('instrument_location');
          nc_ancillary=nc_root.createGroup('ancillary');
          nc_key=nc_root.createGroup('keydata');
          nc_meta=nc_root.createGroup('metadata');
          nc_data=nc_root.createGroup('measurements');

          # Create instrument variables

          instrLatitude=netCDF_NewVariable(nc_instrument,'latitude',float32,('dim1_size',),'degree','latitude of the instrument (positive north)',fillValue,value=float(config.instr_latitude));
          instrLongitude=netCDF_NewVariable(nc_instrument,'longitude',float32,('dim1_size',),'degree','longitude of the instrument (positive east)',fillValue,value=float(config.instr_longitude));
          instrAltitude=netCDF_NewVariable(nc_instrument,'altitude',float32,('dim1_size',),'m','altitude of the instrument above sea level',fillValue,float(config.instr_altitude_asl));
          stationAltitude=netCDF_NewVariable(nc_instrument,'altitude_of_station',float32,('dim1_size',),'m','altitude of the station above sea level',fillValue,float(config.station_altitude_asl));
          # instrAltitudeGround=netCDF_NewVariable(nc_instrument,'altitude_agl',float32,('dim1_size',),'m','altitude of the instrument above ground level',fillValue,float(config.instr_altitude_agl));
          # stationName=netCDF_NewVariable(nc_instrument,'station_name',str,('dim1_size',),'1','station name retrieved from the GEOMS data base','x',value=asarray([config.station_name],dtype='U'));

          # stationName[:]=asarray([config.station_name],dtype='U');

          # Create ancillary variables

          nc_meteo=nc_ancillary.createGroup('meteorological_data');

          meteoTime=netCDF_NewVariable(nc_meteo,'meteo_time',float32,('dim1_size','meteo_time_size'),'hour','meteo measurement time in fractional time',fillValue,tp.meteo_time);
          meteoAlt=netCDF_NewVariable(nc_meteo,'altitude_levels',float32,('meteo_levels_size',),'km','altitudes levels above sea level',fillValue,value=tp.altitude_levels);
          meteoPres=netCDF_NewVariable(nc_meteo,'pressure',float32,('meteo_levels_size','meteo_time_size'),'hPa','pressure profile',fillValue,value=tp.pressure);
          meteoTemp=netCDF_NewVariable(nc_meteo,'temperature',float32,('meteo_levels_size','meteo_time_size'),'K','temperature profile',fillValue,value=tp.temperature);
          
          # Create key data

          if slfSize[0]>0 and slfSize[1]>0:

             nc_slit=nc_key.createGroup('slit_function');

             slitX=netCDF_NewVariable(nc_slit,'slit_function_relative_wavelengths',float32,('slit_dimx',),'nm','relative wavelength grid',fillValue,value=slfX);                                              # relative wavelengths
             slitY=netCDF_NewVariable(nc_slit,'slit_function_measured_wavelengths',float32,('dim1_size','slit_dimy'),'nm','wavelengths at which the slit function has been measured',fillValue,value=slfY);   # measured wavelengths
             slit=netCDF_NewVariable(nc_slit,'slit_function',float32,('slit_dimx','slit_dimy'),'1','measured slit function(s)',fillValue,value=slf);                                                          # measured slit functions
          
          if refSize[0]==detsize and refSize[1]>=2:

             nc_ref=nc_key.createGroup('reference_spectrum');

             refWve=netCDF_NewVariable(nc_ref,'ref_wavelengths',float32,('det_size',),'nm','reference grid',fillValue,value=ref[:,0]);
             refSpe=netCDF_NewVariable(nc_ref,'ref_spectrum',float32,('det_size',),'1','reference spectrum',fillValue,value=ref[:,1]);

          # Create metadata variables
          #   Azimuth should be in [0..360] range

          L1_data.metadata.viewingAzimuth=(L1_data.metadata.viewingAzimuth+viewingAzimuthCorrection)%360.;
          L1_data.metadata.solarAzimuthAngle=(L1_data.metadata.solarAzimuthAngle+solarAzimuthCorrection)%360.;
          
          # --- mandatory fields

          viewingElevation=netCDF_NewVariable(nc_meta,'elevation_viewing_angle',float32,('num_rec',),'degree','viewing elevation angle',fillValue,value=L1_data.metadata.viewingElevation);
          viewingAzimuth=netCDF_NewVariable(nc_meta,'azimuth_viewing_angle',float32,('num_rec',),'degree','viewing azimuth angle 0..360, measured towards the east, from north',fillValue,value=L1_data.metadata.viewingAzimuth);
          solarZenithAngle=netCDF_NewVariable(nc_meta,'solar_zenith_angle',float32,('num_rec',),'degree','solar zenith angle',fillValue,value=L1_data.metadata.solarZenithAngle);
          solarAzimuthAngle=netCDF_NewVariable(nc_meta,'solar_azimuth_angle',float32,('num_rec',),'degree','solar azimuth angle, 0..360, measured towards the east, from north',fillValue,value=L1_data.metadata.solarAzimuthAngle);
          exposureTime=netCDF_NewVariable(nc_meta,'exposure_time',float32,('num_rec',),'sec','exposure time',fillValue,value=L1_data.metadata.exposureTime);
          measureType=netCDF_NewVariable(nc_meta,'measurement_type',int16,('num_rec',),'1','measurement type : 1-offaxis, 2-direct sun, 3-zenith, 7-almucantar, 11-horizon, 12-direct moon',-1,value=L1_data.metadata.measureType);

                             # not necessary anymore longDate=netCDF_NewVariable(nc_meta,'long_datetime',str,('num_rec',),'1','UT date and time in YYYYMMDDhhmmss.ms format','000000000000.00');
                             # fracDay=netCDF_NewVariable(nc_meta,'fractional_day',float32,('num_rec',),'UT calendar day','fractional calendar day');
                             # fracTime=netCDF_NewVariable(nc_meta,'fractional_time',float32,('num_rec',),'UT fractional hours','fractional time');
                             # fracStartTime=netCDF_NewVariable(nc_meta,'fractional_start_time',float32,('num_rec',),'UT fractional hours','fractional start time');
                             # fracEndTime=netCDF_NewVariable(nc_meta,'fractional_end_time',float32,('num_rec',),'UT fractional hours','fractional end time');

          # --- optional fields (for example, for mobile measurements)

                             # --- optional latitude=netCDF_NewVariable(nc_meta,'latitude',float32,('num_rec',),'degrees_north','latitude (positive north)');
                             # --- optional longitude=netCDF_NewVariable(nc_meta,'longitude',float32,('num_rec',),'degrees_east','longitude (positive east)');
                             
          totalAcqTime=netCDF_NewVariable(nc_meta,'total_acquisition_time',float32,('num_rec',),'sec','total acquisition time (the total time the detector is exposed to light to produce the spectrum)',fillValue,value=L1_data.metadata.totalAcqTime);
          totalMeasTime=netCDF_NewVariable(nc_meta,'total_measurement_time',float32,('num_rec',),'sec','total measurement time (should be the time difference in seconds between UTC time start and UTC time stop)',fillValue,value=L1_data.metadata.totalMeasTime);

          # date and time

          for i in range(0,nrecords):
            
             dateCurrent=L1_data.metadata.utTime[i,:];
             dateStart=L1_data.metadata.utStart[i,:];
             dateEnd=L1_data.metadata.utEnd[i,:];

             d1=datetime(dateStart[0],dateStart[1],dateStart[2],dateStart[3],dateStart[4],dateStart[5]);
             d=datetime(dateCurrent[0],dateCurrent[1],dateCurrent[2],dateCurrent[3],dateCurrent[4],dateCurrent[5]);
             d2=datetime(dateEnd[0],dateEnd[1],dateEnd[2],dateEnd[3],dateEnd[4],dateEnd[5]);

             if d1>d:
                d1=d1-timedelta(days=1);
                dateStart[0]=d1.year;
                dateStart[1]=d1.month;
                dateStart[2]=d1.day;

             if d2<d:
                d2=d2+timedelta(days=1);
                dateEnd[0]=d1.year;
                dateEnd[1]=d1.month;
                dateEnd[2]=d1.day;             

             L1_data.metadata.utTime[i,:]=dateCurrent;
             L1_data.metadata.utStart[i,:]=dateStart;
             L1_data.metadata.utEnd[i,:]=dateEnd;

          utTime=netCDF_NewVariable(nc_meta,'datetime',int16,('num_rec','datetime_size'),'1','measurement date and time (UT YYYY,MM,DD,hh,mm,ss,ms)',0,value=L1_data.metadata.utTime);
          utStart=netCDF_NewVariable(nc_meta,'datetime_start',int16,('num_rec','datetime_size'),'1','start date and time (UT YYYY,MM,DD,hh,mm,ss,ms)',0,value=L1_data.metadata.utStart);
          utEnd=netCDF_NewVariable(nc_meta,'datetime_end',int16,('num_rec','datetime_size'),'1','end date and time (UT YYYY,MM,DD,hh,mm,ss,ms)',0,value=L1_data.metadata.utEnd);

          # Create measurements variables

          wavelengths=netCDF_NewVariable(nc_data,'wavelengths',float32,('num_rec','det_size'),'nm','wavelengths',fillValue,value=L1_data.measurements.wavelength);
          spectra=netCDF_NewVariable(nc_data,'radiances',float32,('num_rec','det_size'),'1','spectra (counts number)',fillValue,value=L1_data.measurements.spectra);
          errors=netCDF_NewVariable(nc_data,'radiances_errors',float32,('num_rec','det_size'),'1','errors (counts number)',fillValue);
          pixelQualityFlags=netCDF_NewVariable(nc_data,'radiances_quality_flags',int16,('num_rec','det_size'),'1','pixel quality flags (currently 0 or 1 but probably to complete later)',1);

                             # longDate[:]=asarray(datestr,dtype='U');

          # Close file

          nc_root.close();

    return ncFile;

# Examples of calls
#
# netCDF_L1_Convert('/bira-iasb/projects/FRM4DOAS/tmp/demo/config/level1/XIANGVIS_041_1_ccdeev2netcdf.ini',
#                   '/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Xianghe/Data/spectra/VIS/XIANGVIS/DSC/2018/XIANGVIS_20180122_022.DSC',
#                   '/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Xianghe/Data/spectra/VIS/XIANGVIS/netCDF',
#                   refFile='/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Xianghe/Data/spectra/VIS/XIANGVIS/Calib/XIANGVIS_20100615 _166_244.ref',
#                   file_version='001',
#                   file_format='CCDEEV',
#                   clbFile='/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Xianghe/Data/spectra/VIS/XIANGVIS/Calib/calib_Matlab_2.clb',
#                   commentChar=';',
#                   azimuthCorrection=180.)
#
#
# netCDF_L1_Convert('/bira-iasb/projects/FRM4DOAS/tmp/demo/config/level1/cindi2_004_1_asc2netcdf.ini',
#                   '/bira-iasb/projects/FRM4DOAS/data/spectra/campaigns/cindi2/bira_4/bira_maxdoas_4_vis_spectra_cindi2_20160924_v1.asc',
#                   '/bira-iasb/projects/FRM4DOAS/tmp',
#                   refFile='/bira-iasb/projects/FRM4DOAS/data/results/campaigns/cindi2/tests/bira_4/bira_maxdoas_4_vis_spectra_cindi2_20160924_v1.ref',
#                   file_version='001',
#                   file_format='ASCII');
#
# netCDF_L1_Convert('/bira-iasb/projects/FRM4DOAS/tmp/demo/config/level1/PESVIS_040_1_mfcbira2netcdf.ini',
#                  '/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Antarctica/Envimes/Data/spectra/VIS/PESVIS/MFC-BIRA/Part2/2017/PESVIS_20171206.bin',
#                  '/bira-iasb/projects/FRM4DOAS/tmp',
#                  refFile='/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Antarctica/Envimes/Data/spectra/VIS/Calib/PESVIS_20180214_587.ref',
#                  file_version='001',
#                  file_format='MFC_BIRA',
#                  clbFile='/bira-iasb/projects/DOAS/Ground-based-Activities/GB_Stations/Antarctica/Envimes/Data/spectra/VIS/Calib/ENVIMES_CalibVis.clb',
#                  commentChar=';',
#                  azimuthCorrection=180.);

