
# ------------------------------------------------------------------------------
# MODULE        netCDF_tools
# ------------------------------------------------------------------------------
# PURPOSE       Tools to handle netCDF files
# ------------------------------------------------------------------------------
# AUTHOR        Caroline FAYT
# COPYRIGHT     Royal Belgian Institute for Space Aeronomy
# CREATION DATE January 2018
#
# MODIFIED
#
# ------------------------------------------------------------------------------

import os;
import netCDF4;

from numpy import *

def netCDF_NewVariable(root,varname,vartype,varshape,varunits,vardesc,varfill,value=[]):

    """
    variable=netCDF_NewVariable(root,varname,vartype,varshape,varunits,vardesc,varfill,value=[])

    Create a new variable in a netCDF file.  The name, type, shape, units, description and filling value should be provided.
    The content of the variable can already be passed as argument.

    Parameters
    ----------

    root        : the group id the variable should belong to
    varname     : the name of the variable
    vartype     : the type of the variable
    varshape    : the shape of the variable
    varunits    : the units of the variable
    vardesc     : the description of the variable
    varfill     : the filling value
    value       : optional, the content of the variable

    Returns
    -------

    variable : returned by root.createVariable

    See also
    --------

    netCDF4
    """

    variable=root.createVariable(varname,vartype,varshape,zlib=True,fill_value=varfill)
    variable.units=varunits;
    variable.description=vardesc;

    # todo : should check that the dimensions match

    if size(value)>0:
      variable[:]=value;

    return variable;

def netCDF_GetAttributes(ncFile):

    """
    attrList=netCDF_GetAttributes(ncFile)

    Returns the list of main attributes of a netCDF file as a dictionary

    Parameters
    ----------

    ncFile      : the name of the netCDF file

    Returns
    -------

    attrList : the list of main attributes of the input netCDF file as a dictionary

    See also
    --------

    netCDF4
    """

    functionName="netCDF_getAttributes"
    attrList={};

    if not os.path.isfile(ncFile):
       print('['+functionName+'] '+ncFile+' not found !!!');
    else:

       try:

           keys=[];

           nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');
           attrs=nc.ncattrs();
           for item in attrs:
               keys.append(nc.getncattr(item));
           nc.close();

           attrList=dict(zip(attrs,keys));

       except OSError :
           print('['+functionName+'] '+'Can not open file '+ncFile+' or attributes not found !!!');

    return attrList;

def netCDF_CopyAttributes(ncFileSource,ncFileTarget,attrList):

    """
    netCDF_CopyAttributes(ncFileSource,ncFileTarget,attrList)

    Copy main attributes from one netCDF file to another

    Parameters
    ----------

    ncFileSource   : the name of the source netCDF file
    ncFileTarget   : the name of the target netCDF file
    attrList       : the list of attributes to copy

    See also
    --------

    netCDF4
    """

    functionName="netCDF_copyAttributes"

    if not os.path.isfile(ncFileSource):
       print('['+functionName+'] '+ncFileSource+' not found !!!');
    elif not os.path.isfile(ncFileTarget):
       print('['+functionName+'] '+ncFileTarget+' not found !!!');
    else:

       try:

           sourceAttr=netCDF_GetAttributes(ncFileSource);
           targetAttr=netCDF_GetAttributes(ncFileTarget);

           nc=netCDF4.Dataset(ncFileTarget,'r+',format='NETCDF4');

           for item in attrList:

               if item not in sourceAttr:
                   print('WARNING : attribute '+item+' not in '+ncFileSource);
               else:

                   if (item in targetAttr) and (sourceAttr[item]!=targetAttr[item]):
                       print('WARNING : attribute '+item+' modified (' +targetAttr[item]+' -> '+ sourceAttr[item]+')');

                   nc.setncattr_string(item,sourceAttr[item]);

           nc.close()

       except OSError:

           print('['+functionName+'] '+'ERROR : can not copy attributes from '+ncFileSource+' to '+ncFileTarget);

def netCDF_Datetime(datestr,timestr):

    """
    netCDF_Datetime(datestr,timestr)

    Decompose date and time strings in respectively dd/mm/yyyy and hh:mm:ss.ms formats, in seven fields
    yyyy,mm,dd,hh,mm,ss,ms to save in a netCDF file

    Parameters
    ----------

    datestr   : date string in dd/mm/yyyy format
    timestr   : time string in hh:mm:ss.ms format

    Returns
    -------

    a seven values array with the different date and time fields (yyyy,mm,dd,hh,mm,ss,ms)
    """

    # todo : check the validity of the date

    dt=datestr.split('/')
    ut=timestr.split(':');
    ss=ut[2];
    ms='0';

    if '.' in ss:
      ssms=ss.split('.');

      ss=ssms[0];
      ms=ssms[1];

    return asarray([int16(dt[2]),int16(dt[1]),int16(dt[0]),int16(ut[0]),int16(ut[1]),int16(ss),int16(ms)])          # !!! take care of milliseconds for some formats

def netCDF_ReadQDOAS(qdoasFile):

    """
    netCDF_ReadQDOAS(qdoasFile)

    Read a QDOAS output file in netCDF format

    Parameters
    ----------

    qdoasFile   : the name of a file generated by QDOAS

    Returns
    -------

    a structure similar to the one returned by ASCTAB_Read
    """

    functionName="netCDF_ReadQDOAS"
    data={};

    if not os.path.isfile(qdoasFile):

       print ('['+functionName+'] '+'File ' + qdoasFile + ' do not exist')

    else:

       try:

          # test if it is possible to open file

          nc=netCDF4.Dataset(qdoasFile,'r',format='NETCDF4');

          if len(nc.groups)==0:
             print('['+functionName+'] '+qdoasFile+' is empty !!!');
          elif len(nc.groups)!=1 :
             print('['+functionName+'] '+qdoasFile+' is not a QDOAS file !!!');
          else:

             qdoas_name=list(nc.groups.keys())[0];
             qdoas_group=nc.groups[qdoas_name];

             for qdoas_var in qdoas_group.variables:
                 data[qdoas_var]=asarray(qdoas_group.variables[qdoas_var]);

             # Browse analysis windows

             for analysis_window in qdoas_group.groups:
                 for qdoas_var in qdoas_group[analysis_window].variables:
                     data[analysis_window+'.'+qdoas_var]=asarray(qdoas_group[analysis_window].variables[qdoas_var])

          nc.close();

       except OSError:

           print('['+functionName+'] '+'ERROR : can not read netCDF file '+qdoasFile);

    return data